/**********************************************************/
/*         BRANCHOUT.C                                    */
/*                         Produced by Niimura            */
/*                         Last Modified on '06.7.24      */
/*                                                        */
/**********************************************************/

#include <string.h>
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#define MAXLINE   100
#define MAXBR   10000
#define MAXSEQ   5000
#define MAXNAME   100

int  seqnum, bootnum, total_branch, root, h_option, b_option;
int  total_outgp, outgp[MAXSEQ], branch_order[MAXSEQ], total_inbranch, total_node;
char name[MAXSEQ][MAXNAME];
struct { int    node1, node2, boot, member[MAXSEQ], membernum, child[2];
	 char   mark;
       } branch[MAXBR];
struct { int    branch1, branch2;
	 char   mark;
       } node[MAXBR];

void GetData();
void GetOutgp();
void GetSeparation();
void FindRoot();
void MarkBranch();
void MarkNode();
void SortBranch();
void GetChildren();
void MakeRoot2();
void BranchOut();
void Reverse();
int  intcmp();
void swap();

main(argc, argv)
int   argc;
char *argv[];
{
	FILE *fp, *fp2;
	int  i, j, k;

        if (argc < 3)
        {
		fprintf(stderr,"\n");
		fprintf(stderr,"This program outputs branch information.\n");
		fprintf(stderr,"#Output is similar to 'mklintree.c' with -b option, ");
		fprintf(stderr,"but this does not require SEQDIST_OUTPUT_FILE.\n");
		fprintf(stderr,"%s  LINTRE_OUTPUT_FILE  -o OUTGROUP1 OUTGROUP2...\n", argv[0]);
		fprintf(stderr,"         or\n");
		fprintf(stderr,"%s  LINTRE_OUTPUT_FILE  -f OUTGROUP_LIST_FILE\n", argv[0]);
		fprintf(stderr,"#OUTGROUP_LIST_FILE should contain one number in each line.\n\n");
		fprintf(stderr,"Output:\n");
		fprintf(stderr,"Branch_# I/O/R Child_branch_1 Child_branch_2 Bootstrap ");
		fprintf(stderr,"#_of_genes_below_the_branch Genes_below_the_branch\n");
		fprintf(stderr,"#The last line ('r') indicates the opposite direction of the root branch.\n\n");
		exit(-1);
	}

	if ((fp = fopen(argv[1],"r")) == NULL)
	{
		fprintf(stderr,"Cannot open %s\n",argv[1]);
		exit(-1);
        }
	GetData(fp);

	for (i = 0; i < argc; i++)
	{
		if (strcmp(argv[i],"-f") == 0)
		{
			if ((fp2 = fopen(argv[i+1],"r")) == NULL)
			{
				fprintf(stderr,"Cannot open %s\n",argv[i+1]);
				exit(-1);
			}
			GetOutgp(fp2);
		}
		else if (strcmp(argv[i],"-o") == 0)
		{
			k = 0;
			for (j = i + 1; j < argc; j++)
			{
				if (argv[j][0] == '-')
					break;
				outgp[k] = atoi(argv[j]);
				k++;
			}
			total_outgp = k;
		}
	}
	qsort(outgp, total_outgp, sizeof(int), intcmp);
	GetSeparation();
	FindRoot();
	MarkBranch();
	MarkNode();
	SortBranch();
	GetChildren();
	MakeRoot2();
	BranchOut();
}

void GetData(fp)
FILE *fp;
{
        char   lines[MAXLINE], e[MAXLINE];
	int    i, l, a, b, d;
	double c;

	l = 1;
	i = 0;
	while(fgets(lines,MAXLINE,fp) != NULL)
	{
		if (l == 1)
		{
			if (strlen(lines) > 20)
			{
				sscanf(lines,"%d sequences %d bootstrapping",&a,&b);
				seqnum  = a;
				bootnum = b;
			}
			else
			{
				sscanf(lines,"%d sequences",&a);
				seqnum  = a;
				bootnum = 0;
			}
			total_branch = 2 * seqnum - 3;
			total_node = total_branch + 1;
		}
		else if (l < seqnum + 2)
		{
			sscanf(lines,"%d %s\n",&a,e);
                        strcpy(name[a-1], e);
		}
		else if (l < 3 * seqnum - 1) 
		{
			if (bootnum == 0)
			{
				sscanf(lines,"%d and %d %lf",&a,&b,&c);
				branch[i].boot = 0;
			}
			else
			{
				sscanf(lines,"%d and %d %lf %d",&a,&b,&c,&d);
				branch[i].boot = d;
			}
			branch[i].node1 = a;
			branch[i].node2 = b;
			i++;
		}
		l++;
	}
}

void GetOutgp(fp)
FILE *fp;
{
	char   lines[MAXLINE];
	int    i;

	i = 0;
	while(fgets(lines,MAXLINE,fp) != NULL)
	{
		outgp[i] = atoi(lines);
		i++;
	}
	total_outgp = i;
}

void GetSeparation()
{
	int  i, j, k, flag;
	int  intcmp();

	for (i = 0; i < total_branch; i++)
	{
		if (branch[i].node2 < seqnum + 1)
		{
			branch[i].member[0] = branch[i].node2;
			branch[i].membernum = 1;
		}
		else
		{
			flag = 0;
			for (j = 0; j < i; j++)
			{
				if (branch[j].node1 == branch[i].node2)
				{
					for (k = 0; k < branch[j].membernum; k++)
						branch[i].member[branch[i].membernum + k] = branch[j].member[k];
					branch[i].membernum += branch[j].membernum;
					flag++;
					if (flag == 2)
						break;
				}
			}
		}
	}
	for (i = 0; i < total_branch; i++)
		qsort(branch[i].member, branch[i].membernum, sizeof(int), intcmp);
	return;
}

void FindRoot()
{
	int  i, j, flag;
	int  check_root();

	for (i = 0; i < total_branch; i++)
	{
		if (branch[i].membernum == total_outgp)
		{
			root = check_root(i);
			if (root != -1)
				return;
		}
	}
	for (i = 0; i < total_branch; i++)
	{
		if (branch[i].membernum == seqnum - total_outgp)
		{
			Reverse(i);
			root = check_root(i);
			if (root != -1)
				return;
		}
	}
	printf("Root OUTs are not monophyletic!\n");
	exit(-1);
}

int check_root(int i)
{
	int  j, flag;

	flag = 0;
	for (j = 0; j < total_outgp; j++)
		if (branch[i].member[j] != outgp[j])
		{
			flag = 1;
			break;
		}
	if (flag == 0)
	{
		branch[i].mark = 'R';		/*** Root ***/
		Reverse(i);	/*** The second node should be included in the linearized tree. ***/
		return(i);
	}
	else
		return(-1);
}

void Reverse(int i)
{
	int  j, k, tmp[MAXSEQ];

	k = 0;
	for (j = 1; j < seqnum + 1; j++)
	{
		if (j == branch[i].member[k])
		{
			tmp[j] = 0;
			k++;
		}
		else
			tmp[j] = 1;
	}
	k = 0;
	for (j = 1; j < seqnum + 1; j++)
		if (tmp[j] == 1)
		{
			branch[i].member[k] = j;
			k++;
		}
	branch[i].membernum = k;
	branch[i].member[k] = '\0';
	swap(&branch[i].node1, &branch[i].node2);
}

void MarkBranch()
{
	int  i, inout;
	int  in_out();

	for (i = 0; i < total_branch; i++)
	{
		if (i == root)
			continue;

		inout = in_out(i);
		if (inout == -1)
		{
			Reverse(i);
			in_out(i);
		}
	}
}

int in_out(int i)
{
	int  j, k, out_num;	

	out_num = 0;
	for (j = 0; j < branch[i].membernum; j++)
		for (k = 0; k < total_outgp; k++)
			if (branch[i].member[j] == outgp[k])
			{
				out_num++;
				break;
			}

	if (out_num == branch[i].membernum)
	{
		branch[i].mark = 'O';		/*** Outgroup ***/
		return(0);
	}
	else if (out_num == 0)
	{
		branch[i].mark = 'I';		/*** Used for the linearized tree ***/
		return(0);
	}
	else
		return(-1);
}

void swap(int *x, int *y)
{
	int tmp;

	tmp = *x; *x = *y; *y = tmp;
}

int intcmp(int *a, int *b)
{
        return(*a - *b);
}

void MarkNode()
{
	int  i;

	for (i = 0; i < total_branch; i++)
	{
		if (branch[i].mark == 'O')
		{
			node[branch[i].node1].mark = 'O';
			node[branch[i].node2].mark = 'O';
		}
		else if (branch[i].mark == 'I')
		{
			node[branch[i].node1].mark = 'I';
			node[branch[i].node2].mark = 'I';

			if (node[branch[i].node1].branch1)
				node[branch[i].node1].branch2 = i;
			else
				node[branch[i].node1].branch1 = i;
		}
	}
}

void SortBranch()
{
	int  i, j;
	int  membernum_cmp();

	j = 0;
	for (i = 0; i < total_branch; i++)
		if (branch[i].mark == 'I')
		{
			branch_order[j] = i;
			j++;
		}

	total_inbranch = j;
	qsort(branch_order, total_inbranch, sizeof(int), membernum_cmp);
}

int membernum_cmp(int *a, int *b)
{
	return(branch[*b].membernum - branch[*a].membernum);
}

void GetChildren()
{
	int  i, j, k;

	for (i = 0; i < total_branch; i++)
	{
		if (branch[i].node2 < seqnum + 1)
		{
			branch[i].child[0] = -1;
			branch[i].child[1] = -1;
			continue;
		}
		k = 0;
		for (j = 0; j < total_branch; j++)
		{
			if (branch[i].node2 == branch[j].node1)
			{
				branch[i].child[k] = j;
				k++;
			}
			if (k == 2)
				break;
		}
	}
}

void MakeRoot2()
{
	int  i, j, max, child1, child2, flag;
	
	branch[total_branch].mark = 'r';
	branch[total_branch].membernum = total_outgp;
	for (i = 0; i < total_outgp; i++)
		branch[total_branch].member[i] = outgp[i];
	for (i = 0; i < total_branch; i++)
		if (branch[i].mark == 'R')
			break;
	branch[total_branch].boot = branch[i].boot;
	
	if (total_outgp == 1)
		child1 = child2 = -1;
	else
	{
		max = 0;		
		for (i = 0; i < total_branch; i++)
		{
			if (branch[i].mark == 'O' && branch[i].membernum > max)
			{
				child1 = i;
				max = branch[i].membernum;
			}
		}	
		for (i = 0; i < total_branch; i++)
		{
			if (branch[i].mark == 'O' && branch[i].membernum == total_outgp - branch[child1].membernum)
			{
				flag = 0;
				for (j = 0; j < branch[child1].membernum; j++)
				{					
					if (branch[i].member[0] == branch[child1].member[j])
					{					
						flag = 1;
						break;
					}
				}
				if (flag == 0)
				{
					child2 = i;
					break;
				}
			}
		}
	}	
	if (child1 > child2)
		swap(&child1, &child2);
	branch[total_branch].child[0] = child1;
	branch[total_branch].child[1] = child2;
} 


void BranchOut()
{
	int  i, j;

	for (i = 0; i < total_branch+1; i++)
	{
		printf("%4d  %c  %4d  %4d  %4d  %4d  ",
			i+1, branch[i].mark, branch[i].child[0] + 1, branch[i].child[1] + 1,
			branch[i].boot, branch[i].membernum);
		for (j = 0; j < branch[i].membernum; j++)
			printf("%s ", name[branch[i].member[j] - 1]);
		printf("\n");
	}
}
